@@ -1,10 +1,8 @@
+/* A Bison parser, made by GNU Bison 2.5.  */
 
-/* A Bison parser, made by GNU Bison 2.4.1.  */
-
-/* Skeleton interface for Bison's Yacc-like parsers in C
+/* Bison interface for Yacc-like parsers in C
    
-      Copyright (C) 1984, 1989, 1990, 2000, 2001, 2002, 2003, 2004, 2005, 2006
-   Free Software Foundation, Inc.
+      Copyright (C) 1984, 1989-1990, 2000-2011 Free Software Foundation, Inc.
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -316,215 +314,218 @@
      LC_COLLATE_P = 532,
      LC_CTYPE_P = 533,
      LEADING = 534,
-     LEAST = 535,
-     LEFT = 536,
-     LEVEL = 537,
-     LIKE = 538,
-     LIMIT = 539,
-     LISTEN = 540,
-     LOAD = 541,
-     LOCAL = 542,
-     LOCALTIME = 543,
-     LOCALTIMESTAMP = 544,
-     LOCATION = 545,
-     LOCK_P = 546,
-     MAPPING = 547,
-     MATCH = 548,
-     MAXVALUE = 549,
-     MINUTE_P = 550,
-     MINVALUE = 551,
-     MODE = 552,
-     MONTH_P = 553,
-     MOVE = 554,
-     NAME_P = 555,
-     NAMES = 556,
-     NATIONAL = 557,
-     NATURAL = 558,
-     NCHAR = 559,
-     NEXT = 560,
-     NO = 561,
-     NONE = 562,
-     NOT = 563,
-     NOTHING = 564,
-     NOTIFY = 565,
-     NOTNULL = 566,
-     NOWAIT = 567,
-     NULL_P = 568,
-     NULLIF = 569,
-     NULLS_P = 570,
-     NUMERIC = 571,
-     OBJECT_P = 572,
-     OF = 573,
-     OFF = 574,
-     OFFSET = 575,
-     OIDS = 576,
-     ON = 577,
-     ONLY = 578,
-     OPERATOR = 579,
-     OPTION = 580,
-     OPTIONS = 581,
-     OR = 582,
-     ORDER = 583,
-     OUT_P = 584,
-     OUTER_P = 585,
-     OVER = 586,
-     OVERLAPS = 587,
-     OVERLAY = 588,
-     OWNED = 589,
-     OWNER = 590,
-     PARSER = 591,
-     PARTIAL = 592,
-     PARTITION = 593,
-     PASSING = 594,
-     PASSWORD = 595,
-     PLACING = 596,
-     PLANS = 597,
-     POSITION = 598,
-     PRECEDING = 599,
-     PRECISION = 600,
-     PRESERVE = 601,
-     PREPARE = 602,
-     PREPARED = 603,
-     PRIMARY = 604,
-     PRIOR = 605,
-     PRIVILEGES = 606,
-     PROCEDURAL = 607,
-     PROCEDURE = 608,
-     QUOTE = 609,
-     RANGE = 610,
-     READ = 611,
-     REAL = 612,
-     REASSIGN = 613,
-     RECHECK = 614,
-     RECURSIVE = 615,
-     REF = 616,
-     REFERENCES = 617,
-     REINDEX = 618,
-     RELATIVE_P = 619,
-     RELEASE = 620,
-     RENAME = 621,
-     REPEATABLE = 622,
-     REPLACE = 623,
-     REPLICA = 624,
-     RESET = 625,
-     RESTART = 626,
-     RESTRICT = 627,
-     RETURNING = 628,
-     RETURNS = 629,
-     REVOKE = 630,
-     RIGHT = 631,
-     ROLE = 632,
-     ROLLBACK = 633,
-     ROW = 634,
-     ROWS = 635,
-     RULE = 636,
-     SAVEPOINT = 637,
-     SCHEMA = 638,
-     SCROLL = 639,
-     SEARCH = 640,
-     SECOND_P = 641,
-     SECURITY = 642,
-     SELECT = 643,
-     SEQUENCE = 644,
-     SEQUENCES = 645,
-     SERIALIZABLE = 646,
-     SERVER = 647,
-     SESSION = 648,
-     SESSION_USER = 649,
-     SET = 650,
-     SETOF = 651,
-     SHARE = 652,
-     SHOW = 653,
-     SIMILAR = 654,
-     SIMPLE = 655,
-     SMALLINT = 656,
-     SOME = 657,
-     STABLE = 658,
-     STANDALONE_P = 659,
-     START = 660,
-     STATEMENT = 661,
-     STATISTICS = 662,
-     STDIN = 663,
-     STDOUT = 664,
-     STORAGE = 665,
-     STRICT_P = 666,
-     STRIP_P = 667,
-     SUBSTRING = 668,
-     SYMMETRIC = 669,
-     SYSID = 670,
-     SYSTEM_P = 671,
-     TABLE = 672,
-     TABLES = 673,
-     TABLESPACE = 674,
-     TEMP = 675,
-     TEMPLATE = 676,
-     TEMPORARY = 677,
-     TEXT_P = 678,
-     THEN = 679,
-     TIME = 680,
-     TIMESTAMP = 681,
-     TO = 682,
-     TRAILING = 683,
-     TRANSACTION = 684,
-     TREAT = 685,
-     TRIGGER = 686,
-     TRIM = 687,
-     TRUE_P = 688,
-     TRUNCATE = 689,
-     TRUSTED = 690,
-     TYPE_P = 691,
-     UNBOUNDED = 692,
-     UNCOMMITTED = 693,
-     UNENCRYPTED = 694,
-     UNION = 695,
-     UNIQUE = 696,
-     UNKNOWN = 697,
-     UNLISTEN = 698,
-     UNLOGGED = 699,
-     UNTIL = 700,
-     UPDATE = 701,
-     USER = 702,
-     USING = 703,
-     VACUUM = 704,
-     VALID = 705,
-     VALIDATE = 706,
-     VALIDATOR = 707,
-     VALUE_P = 708,
-     VALUES = 709,
-     VARCHAR = 710,
-     VARIADIC = 711,
-     VARYING = 712,
-     VERBOSE = 713,
-     VERSION_P = 714,
-     VIEW = 715,
-     VOLATILE = 716,
-     WHEN = 717,
-     WHERE = 718,
-     WHITESPACE_P = 719,
-     WINDOW = 720,
-     WITH = 721,
-     WITHOUT = 722,
-     WORK = 723,
-     WRAPPER = 724,
-     WRITE = 725,
-     XML_P = 726,
-     XMLATTRIBUTES = 727,
-     XMLCONCAT = 728,
-     XMLELEMENT = 729,
-     XMLEXISTS = 730,
-     XMLFOREST = 731,
-     XMLPARSE = 732,
-     XMLPI = 733,
-     XMLROOT = 734,
-     XMLSERIALIZE = 735,
-     YEAR_P = 736,
-     YES_P = 737,
-     ZONE = 738,
-     NULLS_FIRST = 739,
-     NULLS_LAST = 740,
-     WITH_TIME = 741,
-     POSTFIXOP = 742,
-     UMINUS = 743
+     LEAKPROOF = 535,
+     LEAST = 536,
+     LEFT = 537,
+     LEVEL = 538,
+     LIKE = 539,
+     LIMIT = 540,
+     LISTEN = 541,
+     LOAD = 542,
+     LOCAL = 543,
+     LOCALTIME = 544,
+     LOCALTIMESTAMP = 545,
+     LOCATION = 546,
+     LOCK_P = 547,
+     MAPPING = 548,
+     MATCH = 549,
+     MAXVALUE = 550,
+     MINUTE_P = 551,
+     MINVALUE = 552,
+     MODE = 553,
+     MONTH_P = 554,
+     MOVE = 555,
+     NAME_P = 556,
+     NAMES = 557,
+     NATIONAL = 558,
+     NATURAL = 559,
+     NCHAR = 560,
+     NEXT = 561,
+     NO = 562,
+     NONE = 563,
+     NOT = 564,
+     NOTHING = 565,
+     NOTIFY = 566,
+     NOTNULL = 567,
+     NOWAIT = 568,
+     NULL_P = 569,
+     NULLIF = 570,
+     NULLS_P = 571,
+     NUMERIC = 572,
+     OBJECT_P = 573,
+     OF = 574,
+     OFF = 575,
+     OFFSET = 576,
+     OIDS = 577,
+     ON = 578,
+     ONLY = 579,
+     OPERATOR = 580,
+     OPTION = 581,
+     OPTIONS = 582,
+     OR = 583,
+     ORDER = 584,
+     OUT_P = 585,
+     OUTER_P = 586,
+     OVER = 587,
+     OVERLAPS = 588,
+     OVERLAY = 589,
+     OWNED = 590,
+     OWNER = 591,
+     PARSER = 592,
+     PARTIAL = 593,
+     PARTITION = 594,
+     PASSING = 595,
+     PASSWORD = 596,
+     PLACING = 597,
+     PLANS = 598,
+     POSITION = 599,
+     PRECEDING = 600,
+     PRECISION = 601,
+     PRESERVE = 602,
+     PREPARE = 603,
+     PREPARED = 604,
+     PRIMARY = 605,
+     PRIOR = 606,
+     PRIVILEGES = 607,
+     PROCEDURAL = 608,
+     PROCEDURE = 609,
+     QUOTE = 610,
+     RANGE = 611,
+     READ = 612,
+     REAL = 613,
+     REASSIGN = 614,
+     RECHECK = 615,
+     RECURSIVE = 616,
+     REF = 617,
+     REFERENCES = 618,
+     REINDEX = 619,
+     RELATIVE_P = 620,
+     RELEASE = 621,
+     RENAME = 622,
+     REPEATABLE = 623,
+     REPLACE = 624,
+     REPLICA = 625,
+     RESET = 626,
+     RESTART = 627,
+     RESTRICT = 628,
+     RETURNING = 629,
+     RETURNS = 630,
+     REVOKE = 631,
+     RIGHT = 632,
+     ROLE = 633,
+     ROLLBACK = 634,
+     ROW = 635,
+     ROWS = 636,
+     RULE = 637,
+     SAVEPOINT = 638,
+     SCHEMA = 639,
+     SCROLL = 640,
+     SEARCH = 641,
+     SECOND_P = 642,
+     SECURITY = 643,
+     SELECT = 644,
+     SEQUENCE = 645,
+     SEQUENCES = 646,
+     SERIALIZABLE = 647,
+     SERVER = 648,
+     SESSION = 649,
+     SESSION_USER = 650,
+     SET = 651,
+     SETOF = 652,
+     SHARE = 653,
+     SHOW = 654,
+     SIMILAR = 655,
+     SIMPLE = 656,
+     SMALLINT = 657,
+     SNAPSHOT = 658,
+     SOME = 659,
+     STABLE = 660,
+     STANDALONE_P = 661,
+     START = 662,
+     STATEMENT = 663,
+     STATISTICS = 664,
+     STDIN = 665,
+     STDOUT = 666,
+     STORAGE = 667,
+     STRICT_P = 668,
+     STRIP_P = 669,
+     SUBSTRING = 670,
+     SYMMETRIC = 671,
+     SYSID = 672,
+     SYSTEM_P = 673,
+     TABLE = 674,
+     TABLES = 675,
+     TABLESPACE = 676,
+     TEMP = 677,
+     TEMPLATE = 678,
+     TEMPORARY = 679,
+     TEXT_P = 680,
+     THEN = 681,
+     TIME = 682,
+     TIMESTAMP = 683,
+     TO = 684,
+     TRAILING = 685,
+     TRANSACTION = 686,
+     TREAT = 687,
+     TRIGGER = 688,
+     TRIM = 689,
+     TRUE_P = 690,
+     TRUNCATE = 691,
+     TRUSTED = 692,
+     TYPE_P = 693,
+     TYPES_P = 694,
+     UNBOUNDED = 695,
+     UNCOMMITTED = 696,
+     UNENCRYPTED = 697,
+     UNION = 698,
+     UNIQUE = 699,
+     UNKNOWN = 700,
+     UNLISTEN = 701,
+     UNLOGGED = 702,
+     UNTIL = 703,
+     UPDATE = 704,
+     USER = 705,
+     USING = 706,
+     VACUUM = 707,
+     VALID = 708,
+     VALIDATE = 709,
+     VALIDATOR = 710,
+     VALUE_P = 711,
+     VALUES = 712,
+     VARCHAR = 713,
+     VARIADIC = 714,
+     VARYING = 715,
+     VERBOSE = 716,
+     VERSION_P = 717,
+     VIEW = 718,
+     VOLATILE = 719,
+     WHEN = 720,
+     WHERE = 721,
+     WHITESPACE_P = 722,
+     WINDOW = 723,
+     WITH = 724,
+     WITHOUT = 725,
+     WORK = 726,
+     WRAPPER = 727,
+     WRITE = 728,
+     XML_P = 729,
+     XMLATTRIBUTES = 730,
+     XMLCONCAT = 731,
+     XMLELEMENT = 732,
+     XMLEXISTS = 733,
+     XMLFOREST = 734,
+     XMLPARSE = 735,
+     XMLPI = 736,
+     XMLROOT = 737,
+     XMLSERIALIZE = 738,
+     YEAR_P = 739,
+     YES_P = 740,
+     ZONE = 741,
+     NULLS_FIRST = 742,
+     NULLS_LAST = 743,
+     WITH_TIME = 744,
+     POSTFIXOP = 745,
+     UMINUS = 746
    };
 #endif
 
@@ -534,12 +535,12 @@
 typedef union YYSTYPE
 {
 
-/* Line 1676 of yacc.c  */
-#line 463 "preproc.y"
+/* Line 2068 of yacc.c  */
+#line 564 "preproc.y"
 
 	double	dval;
 	char	*str;
-	int     ival;
+	int		ival;
 	struct	when		action;
 	struct	index		index;
 	int		tagname;
@@ -552,8 +553,8 @@ typedef union YYSTYPE
 
 
 
-/* Line 1676 of yacc.c  */
-#line 557 "preproc.h"
+/* Line 2068 of yacc.c  */
+#line 558 "preproc.h"
 } YYSTYPE;
 # define YYSTYPE_IS_TRIVIAL 1
 # define yystype YYSTYPE /* obsolescent; will be withdrawn */
