@@ -443,23 +443,6 @@ LEFT JOIN (
 ) AS d ON (a.f1 = d.f1)
 WHERE d.f1 IS NULL;
 
---
--- regression test for proper handling of outer joins within antijoins
---
-
-create temp table tt4x(c1 int, c2 int, c3 int);
-
-explain (costs off)
-select * from tt4x t1
-where not exists (
-  select 1 from tt4x t2
-    left join tt4x t3 on t2.c3 = t3.c1
-    left join ( select t5.c1 as c1
-                from tt4x t4 left join tt4x t5 on t4.c2 = t5.c1
-              ) a1 on t3.c2 = a1.c1
-  where t1.c1 = t2.c2
-);
-
 --
 -- regression test for problems of the sort depicted in bug #3494
 --
@@ -677,6 +660,64 @@ SELECT qq, unique1
   USING (qq)
   INNER JOIN tenk1 c ON qq = unique2;
 
+--
+-- nested nestloops can require nested PlaceHolderVars
+--
+
+create temp table nt1 (
+  id int primary key,
+  a1 boolean,
+  a2 boolean
+);
+create temp table nt2 (
+  id int primary key,
+  nt1_id int,
+  b1 boolean,
+  b2 boolean,
+  foreign key (nt1_id) references nt1(id)
+);
+create temp table nt3 (
+  id int primary key,
+  nt2_id int,
+  c1 boolean,
+  foreign key (nt2_id) references nt2(id)
+);
+
+insert into nt1 values (1,true,true);
+insert into nt1 values (2,true,false);
+insert into nt1 values (3,false,false);
+insert into nt2 values (1,1,true,true);
+insert into nt2 values (2,2,true,false);
+insert into nt2 values (3,3,false,false);
+insert into nt3 values (1,1,true);
+insert into nt3 values (2,2,false);
+insert into nt3 values (3,3,true);
+
+explain (costs off)
+select nt3.id
+from nt3 as nt3
+  left join
+    (select nt2.*, (nt2.b1 and ss1.a3) AS b3
+     from nt2 as nt2
+       left join
+         (select nt1.*, (nt1.id is not null) as a3 from nt1) as ss1
+         on ss1.id = nt2.nt1_id
+    ) as ss2
+    on ss2.id = nt3.nt2_id
+where nt3.id = 1 and ss2.b3;
+
+select nt3.id
+from nt3 as nt3
+  left join
+    (select nt2.*, (nt2.b1 and ss1.a3) AS b3
+     from nt2 as nt2
+       left join
+         (select nt1.*, (nt1.id is not null) as a3 from nt1) as ss1
+         on ss1.id = nt2.nt1_id
+    ) as ss2
+    on ss2.id = nt3.nt2_id
+where nt3.id = 1 and ss2.b3;
+
 --
 -- test case where a PlaceHolderVar is propagated into a subquery
 --
@@ -704,6 +745,83 @@ order by 1,2;
 select * from int4_tbl a full join int4_tbl b on true;
 select * from int4_tbl a full join int4_tbl b on false;
 
+--
+-- test for ability to use a cartesian join when necessary
+--
+
+explain (costs off)
+select * from
+  tenk1 join int4_tbl on f1 = twothousand,
+  int4(sin(1)) q1,
+  int4(sin(0)) q2
+where q1 = thousand or q2 = thousand;
+
+explain (costs off)
+select * from
+  tenk1 join int4_tbl on f1 = twothousand,
+  int4(sin(1)) q1,
+  int4(sin(0)) q2
+where thousand = (q1 + q2);
+
+--
+-- test placement of movable quals in a parameterized join tree
+--
+
+explain (costs off)
+select * from tenk1 t1 left join
+  (tenk1 t2 join tenk1 t3 on t2.thousand = t3.unique2)
+  on t1.hundred = t2.hundred and t1.ten = t3.ten
+where t1.unique1 = 1;
+
+explain (costs off)
+select * from tenk1 t1 left join
+  (tenk1 t2 join tenk1 t3 on t2.thousand = t3.unique2)
+  on t1.hundred = t2.hundred and t1.ten + t2.ten = t3.ten
+where t1.unique1 = 1;
+
+explain (costs off)
+select count(*) from
+  tenk1 a join tenk1 b on a.unique1 = b.unique2
+  left join tenk1 c on a.unique2 = b.unique1 and c.thousand = a.thousand
+  join int4_tbl on b.thousand = f1;
+
+select count(*) from
+  tenk1 a join tenk1 b on a.unique1 = b.unique2
+  left join tenk1 c on a.unique2 = b.unique1 and c.thousand = a.thousand
+  join int4_tbl on b.thousand = f1;
+
+explain (costs off)
+select b.unique1 from
+  tenk1 a join tenk1 b on a.unique1 = b.unique2
+  left join tenk1 c on b.unique1 = 42 and c.thousand = a.thousand
+  join int4_tbl i1 on b.thousand = f1
+  right join int4_tbl i2 on i2.f1 = b.tenthous
+  order by 1;
+
+select b.unique1 from
+  tenk1 a join tenk1 b on a.unique1 = b.unique2
+  left join tenk1 c on b.unique1 = 42 and c.thousand = a.thousand
+  join int4_tbl i1 on b.thousand = f1
+  right join int4_tbl i2 on i2.f1 = b.tenthous
+  order by 1;
+
+explain (costs off)
+select * from
+(
+  select unique1, q1, coalesce(unique1, -1) + q1 as fault
+  from int8_tbl left join tenk1 on (q2 = unique2)
+) ss
+where fault = 122
+order by fault;
+
+select * from
+(
+  select unique1, q1, coalesce(unique1, -1) + q1 as fault
+  from int8_tbl left join tenk1 on (q2 = unique2)
+) ss
+where fault = 122
+order by fault;
+
 --
 -- test handling of potential equivalence clauses above outer joins
 --
@@ -726,6 +844,19 @@ select f1, unique2, case when unique2 is null then f1 else 0 end
   from int4_tbl a left join tenk1 b on f1 = unique2
   where (case when unique2 is null then f1 else 0 end) = 0;
 
+--
+-- another case with equivalence clauses above outer joins (bug #8591)
+--
+
+explain (costs off)
+select a.unique1, b.unique1, c.unique1, coalesce(b.twothousand, a.twothousand)
+  from tenk1 a left join tenk1 b on b.thousand = a.unique1                        left join tenk1 c on c.unique2 = coalesce(b.twothousand, a.twothousand)
+  where a.unique2 = 5530 and coalesce(b.twothousand, a.twothousand) = 44;
+
+select a.unique1, b.unique1, c.unique1, coalesce(b.twothousand, a.twothousand)
+  from tenk1 a left join tenk1 b on b.thousand = a.unique1                        left join tenk1 c on c.unique2 = coalesce(b.twothousand, a.twothousand)
+  where a.unique2 = 5530 and coalesce(b.twothousand, a.twothousand) = 44;
+
 --
 -- check handling of join aliases when flattening multiple levels of subquery
 --
@@ -757,100 +888,6 @@ left join
   ) foo3
 using (join_key);
 
---
--- test successful handling of nested outer joins with degenerate join quals
---
-
-explain (verbose, costs off)
-select t1.* from
-  text_tbl t1
-  left join (select *, '***'::text as d1 from int8_tbl i8b1) b1
-    left join int8_tbl i8
-      left join (select *, null::int as d2 from int8_tbl i8b2) b2
-      on (i8.q1 = b2.q1)
-    on (b2.d2 = b1.q2)
-  on (t1.f1 = b1.d1)
-  left join int4_tbl i4
-  on (i8.q2 = i4.f1);
-
-select t1.* from
-  text_tbl t1
-  left join (select *, '***'::text as d1 from int8_tbl i8b1) b1
-    left join int8_tbl i8
-      left join (select *, null::int as d2 from int8_tbl i8b2) b2
-      on (i8.q1 = b2.q1)
-    on (b2.d2 = b1.q2)
-  on (t1.f1 = b1.d1)
-  left join int4_tbl i4
-  on (i8.q2 = i4.f1);
-
-explain (verbose, costs off)
-select t1.* from
-  text_tbl t1
-  left join (select *, '***'::text as d1 from int8_tbl i8b1) b1
-    left join int8_tbl i8
-      left join (select *, null::int as d2 from int8_tbl i8b2, int4_tbl i4b2) b2
-      on (i8.q1 = b2.q1)
-    on (b2.d2 = b1.q2)
-  on (t1.f1 = b1.d1)
-  left join int4_tbl i4
-  on (i8.q2 = i4.f1);
-
-select t1.* from
-  text_tbl t1
-  left join (select *, '***'::text as d1 from int8_tbl i8b1) b1
-    left join int8_tbl i8
-      left join (select *, null::int as d2 from int8_tbl i8b2, int4_tbl i4b2) b2
-      on (i8.q1 = b2.q1)
-    on (b2.d2 = b1.q2)
-  on (t1.f1 = b1.d1)
-  left join int4_tbl i4
-  on (i8.q2 = i4.f1);
-
-explain (verbose, costs off)
-select t1.* from
-  text_tbl t1
-  left join (select *, '***'::text as d1 from int8_tbl i8b1) b1
-    left join int8_tbl i8
-      left join (select *, null::int as d2 from int8_tbl i8b2, int4_tbl i4b2
-                 where q1 = f1) b2
-      on (i8.q1 = b2.q1)
-    on (b2.d2 = b1.q2)
-  on (t1.f1 = b1.d1)
-  left join int4_tbl i4
-  on (i8.q2 = i4.f1);
-
-select t1.* from
-  text_tbl t1
-  left join (select *, '***'::text as d1 from int8_tbl i8b1) b1
-    left join int8_tbl i8
-      left join (select *, null::int as d2 from int8_tbl i8b2, int4_tbl i4b2
-                 where q1 = f1) b2
-      on (i8.q1 = b2.q1)
-    on (b2.d2 = b1.q2)
-  on (t1.f1 = b1.d1)
-  left join int4_tbl i4
-  on (i8.q2 = i4.f1);
-
-explain (verbose, costs off)
-select * from
-  text_tbl t1
-  inner join int8_tbl i8
-  on i8.q2 = 456
-  right join text_tbl t2
-  on t1.f1 = 'doh!'
-  left join int4_tbl i4
-  on i8.q1 = i4.f1;
-
-select * from
-  text_tbl t1
-  inner join int8_tbl i8
-  on i8.q2 = 456
-  right join text_tbl t2
-  on t1.f1 = 'doh!'
-  left join int4_tbl i4
-  on i8.q1 = i4.f1;
-
 --
 -- test ability to push constants through outer join clauses
 --
@@ -861,26 +898,6 @@ explain (costs off)
 explain (costs off)
   select * from tenk1 a full join tenk1 b using(unique2) where unique2 = 42;
 
---
--- test that quals attached to an outer join have correct semantics,
--- specifically that they don't re-use expressions computed below the join;
--- we force a mergejoin so that coalesce(b.q1, 1) appears as a join input
---
-
-set enable_hashjoin to off;
-set enable_nestloop to off;
-
-explain (verbose, costs off)
-  select a.q2, b.q1
-    from int8_tbl a left join int8_tbl b on a.q2 = coalesce(b.q1, 1)
-    where coalesce(b.q1, 1) > 0;
-select a.q2, b.q1
-  from int8_tbl a left join int8_tbl b on a.q2 = coalesce(b.q1, 1)
-  where coalesce(b.q1, 1) > 0;
-
-reset enable_hashjoin;
-reset enable_nestloop;
-
 --
 -- test join removal
 --
